@@ -14,13 +14,14 @@ use socket2::{Domain, Protocol, SockAddr, Socket, Type};
 use thiserror::Error;
 
 pub mod conversion_utils;
+mod dns_packet_receiver;
 mod oneshot_resolver;
 mod resource_record_manager;
 mod service_discovery;
 mod simple_responder;
 
 pub use oneshot_resolver::OneShotMdnsResolver;
-pub use service_discovery::ServiceDiscovery;
+pub use service_discovery::{InstanceInformation, ServiceDiscovery};
 pub use simple_responder::SimpleMdnsResponder;
 
 const UNICAST_RESPONSE: bool = cfg!(not(test));
